use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::thread::sleep;
use std::time::Duration;
use tokio::runtime::Runtime;
use actix_web::{dev::ServerHandle, middleware, rt, web, HttpRequest, HttpServer};
use actix_web::App as WebApp;
use std::{sync::mpsc, thread, time};

use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;

pub trait UserApp {
    fn get_pkg_name(&self) -> String;
    fn init(&mut self);
    fn update(&mut self);
    fn uninit(&mut self);
}

#[derive(SharedFromSelf)]
pub struct App {
    _self_wp: Weak::<RefCell::<Self>>,
    zone_id: i32,
    node_id: i32,
    instance_id: i32,
    tokio_rt: Runtime,
    user_app: Option<Rc::<RefCell::<dyn UserApp>>>,
}

async fn index(req: HttpRequest) -> &'static str {
    "Hello world!asaaa"
}

async fn run_app(tx: mpsc::Sender<ServerHandle>) -> std::io::Result<()> {
    let server = HttpServer::new(|| {
        WebApp::new()
            .service(web::resource("/index.html").to(|| async { "Hello world!" }))
            .service(web::resource("/").to(index))
    })
    .bind(("127.0.0.1", 8080))?
    .workers(2)
    .run();

    let _ = tx.send(server.handle());

    server.await
}

impl App {

    pub fn instance() -> Rc::<RefCell::<Self>> {
        static mut _INSTANCE: Option<Rc::<RefCell::<App>>> = None;
        unsafe {
            match &_INSTANCE {
                Some(v) => {
                    return v.clone();
                }
                None => {
                    _INSTANCE = Some(Rc::new(RefCell::new(
                        App{
                            zone_id: 0,
                            node_id: 0,
                            instance_id: 0,
                            tokio_rt: Runtime::new().unwrap(),
                            user_app: None,
                            _self_wp: Default::default(),
                        }
                    )));
                    _INSTANCE.clone().unwrap().borrow_mut()._self_wp =
                        Rc::downgrade(&_INSTANCE.clone().unwrap());
                }
            }
            return _INSTANCE.clone().unwrap().clone();
        }
    }

    pub fn init(&mut self, user_app: Rc::<RefCell::<dyn UserApp>>) {
        //self.tokio_rt.enter();
        self.init_http_server();
        self.user_app = Some(user_app);
        crate::Timer::instance().borrow_mut().init();
        self.user_app.as_ref().unwrap().borrow_mut().init();
    }

    pub fn uninit(&mut self) {
        self.user_app.as_ref().unwrap().borrow_mut().uninit();
        crate::Timer::instance().borrow_mut().uninit();
    }

    pub fn run(&mut self) {
        loop {
            crate::Timer::instance().borrow_mut().update();
            sleep(Duration::from_millis(1));
        }
    }

    pub fn get_pkg_name(&self) -> String {
        return self.user_app.as_ref().unwrap().borrow().get_pkg_name();
    }

    pub fn new_uuid(&self) -> i64 {
        return 0;
    }

    pub fn get_instance_id(&self) -> i32 {
        return self.instance_id;
    }

    pub fn get_zone_id(&self) -> i32 {
        return self.zone_id;
    }

    pub fn get_node_id(&self) -> i32 {
        return self.node_id;
    }

    pub fn has_flag(&self) -> bool {
        return false;
    }

    pub fn get_now_second(&self) -> i64 {
        return 0;
    }

    pub fn get_now_millis(&self) -> i64 {
        return 0;
    }

    pub fn get_time_offset(&self) -> i64 {
        return 0;
    }

    fn init_http_server(&mut self) {
        let (tx, _) = mpsc::channel();

        thread::spawn(move || {
            let server_future = run_app(tx);
            rt::System::new().block_on(server_future)
        });
    }

}
